@@ -1,21 +1,21 @@
 from __future__ import absolute_import, division, print_function
 
+import os
+import warnings
 import traceback
 from functools import wraps
 
+from glue.core.data import Subset
 from glue.external.six import string_types
 from glue.core.session import Session
-from glue.core.edit_subset_mode import EditSubsetMode
 from glue.core.hub import HubListener
-from glue.core import Data, Subset
-from glue.core import command
+from glue.core import BaseData
 from glue.core.data_factories import load_data
 from glue.core.data_collection import DataCollection
 from glue.config import settings
-from glue.utils import as_list, PropertySetMixin
 
 
-__all__ = ['Application', 'ViewerBase']
+__all__ = ['Application']
 
 
 def catch_error(msg):
@@ -35,7 +35,7 @@ def catch_error(msg):
 
 def as_flat_data_list(data):
     datasets = []
-    if isinstance(data, Data):
+    if isinstance(data, BaseData):
         datasets.append(data)
     else:
         for d in data:
@@ -55,10 +55,9 @@ class Application(HubListener):
             self._session = Session(data_collection=self._data,
                                     application=self)
 
-        EditSubsetMode().data_collection = self._data
         self._hub = self._session.hub
         self._cmds = self._session.command_stack
-        self._cmds.add_callback(lambda x: self._update_undo_redo_enabled())
+        self._cmds.add_callback(self._update_undo_redo_enabled)
 
         self._settings = {}
         for key, value, validator in settings:
@@ -72,7 +71,7 @@ class Application(HubListener):
     def data_collection(self):
         return self.session.data_collection
 
-    def new_data_viewer(self, viewer_class, data=None):
+    def new_data_viewer(self, viewer_class, data=None, state=None):
         """
         Create a new data viewer, add it to the UI,
         and populate with data
@@ -80,19 +79,26 @@ class Application(HubListener):
         if viewer_class is None:
             return
 
-        c = viewer_class(self._session)
+        if state is not None:
+            c = viewer_class(self._session, state=state)
+        else:
+            c = viewer_class(self._session)
         c.register_to_hub(self._session.hub)
 
-        if data and not c.add_data(data):
-            c.close(warn=False)
-            return
+        if data is not None:
+            if isinstance(data, BaseData):
+                result = c.add_data(data)
+            elif isinstance(data, Subset):
+                result = c.add_subset(data)
+            if not result:
+                c.close(warn=False)
+                return
 
         self.add_widget(c)
-        c.show()
         return c
 
     @catch_error("Failed to save session")
-    def save_session(self, path, include_data=False):
+    def save_session(self, path, include_data=False, absolute_paths=True):
         """ Save the data collection and hub to file.
 
         Can be restored via restore_session
@@ -100,9 +106,25 @@ class Application(HubListener):
         Note: Saving of client is not currently supported. Thus,
         restoring this session will lose all current viz windows
         """
+
         from glue.core.state import GlueSerializer
-        gs = GlueSerializer(self, include_data=include_data)
-        state = gs.dumps(indent=2)
+        gs = GlueSerializer(self,
+                            include_data=include_data,
+                            absolute_paths=absolute_paths)
+
+        # In case relative paths are needed in the session file, we do the
+        # serialization while setting the current directory to the directory
+        # in which the session file will be saved so that the relative paths
+        # are relative to the session file, not the current working directory.
+        start_dir = os.path.abspath('.')
+        session_dir = os.path.dirname(path) or '.'
+
+        try:
+            os.chdir(session_dir)
+            state = gs.dumps(indent=2)
+        finally:
+            os.chdir(start_dir)
+
         with open(path, 'w') as out:
             out.write(state)
 
@@ -123,19 +145,21 @@ class Application(HubListener):
         """
         from glue.core.state import GlueUnSerializer
 
-        with open(path) as infile:
-            state = GlueUnSerializer.load(infile)
+        # In case relative paths are needed in the session file, we do the
+        # loading while setting the current directory to the directory
+        # in which the session file is so that relative paths are interpreted
+        # as relative to the session file.
+        start_dir = os.path.abspath('.')
+        session_dir = os.path.dirname(path) or '.'
+        session_file = os.path.basename(path)
 
-        return state.object('__main__')
-
-    def new_tab(self):
-        raise NotImplementedError()
-
-    def add_widget(self, widget, label=None, tab=None):
-        raise NotImplementedError()
-
-    def close_tab(self):
-        raise NotImplementedError()
+        try:
+            os.chdir(session_dir)
+            with open(session_file) as infile:
+                state = GlueUnSerializer.load(infile)
+            return state.object('__main__')
+        finally:
+            os.chdir(start_dir)
 
     def get_setting(self, key):
         """
@@ -160,7 +184,7 @@ class Application(HubListener):
             yield key, value
 
     @catch_error("Could not load data")
-    def load_data(self, paths, skip_merge=False, auto_merge=False):
+    def load_data(self, paths, auto_merge=False, **kwargs):
         """
         Given a path to a file, load the file as a Data object and add it to
         the current session.
@@ -174,13 +198,12 @@ class Application(HubListener):
         datasets = []
         for path in paths:
             result = load_data(path)
-            if isinstance(result, Data):
+            if isinstance(result, BaseData):
                 datasets.append(result)
             else:
                 datasets.extend(result)
 
-        self.add_datasets(self.data_collection, datasets,
-                          skip_merge=skip_merge, auto_merge=auto_merge)
+        self.add_datasets(datasets, auto_merge=auto_merge, **kwargs)
 
         if len(datasets) == 1:
             return datasets[0]
@@ -202,7 +225,10 @@ class Application(HubListener):
         datasets = []
 
         for path in args:
-            datasets.append(load_data(path))
+            if isinstance(path, BaseData):
+                datasets.append(path)
+            else:
+                datasets.append(load_data(path))
 
         links = kwargs.pop('links', None)
 
@@ -211,7 +237,7 @@ class Application(HubListener):
         for label, data in kwargs.items():
             datasets.extend(parse_data(data, label))
 
-        self.add_datasets(self.data_collection, datasets)
+        self.add_datasets(datasets)
 
         if links is not None:
             self.data_collection.add_link(parse_links(self.data_collection, links))
@@ -249,56 +275,56 @@ class Application(HubListener):
     def _update_undo_redo_enabled(self):
         raise NotImplementedError()
 
-    @classmethod
-    def add_datasets(cls, data_collection, datasets, skip_merge=False, auto_merge=False):
-        """ Utility method to interactively add datasets to a
-        data_collection
+    def add_datasets(self, *args, **kwargs):
+        """
+        Utility method to interactively add datasets to the data_collection.
 
         Parameters
         ----------
-        data_collection : :class:`~glue.core.data_collection.DataCollection`
         datasets : :class:`~glue.core.data.Data` or list of Data
-            One or more :class:`~glue.core.data.Data` instances
+            One or more :class:`~glue.core.data.Data` instances.
 
-        Adds datasets to the collection
+        Adds datasets to the collection in the application.
         """
 
+        if isinstance(args[0], DataCollection):
+            warnings.warn("Calling add_datasets with an explicit data "
+                          "collection is now deprecated. You should now call "
+                          "app.add_datasets(datasets).", UserWarning)
+            data_collection, datasets = args
+        else:
+            data_collection = self.data_collection
+            datasets = args[0]
+
+        if "skip_merge" in kwargs:
+            warnings.warn("The skip_merge= argument now no longer has any "
+                          "effect and is deprecated, since no merging is done "
+                          "by default", UserWarning)
+
+        auto_merge = kwargs.pop('auto_merge', False)
+
         datasets = as_flat_data_list(datasets)
         data_collection.extend(datasets)
 
-        # We now check whether any of the datasets can be merged. We need to
-        # make sure that datasets are only ever shown once, as we don't want
-        # to repeat the menu multiple times.
+        # We now automatically merge the datasets if requested. However, we only
+        # do this if all the datasets have the same shape to avoid confusion.
+        if auto_merge and len(datasets) > 1:
 
-        suggested = []
+            reference_shape = datasets[0].shape
 
-        for data in datasets:
-
-            # If the data was already suggested, we skip over it
-            if data in suggested:
-                continue
-
-            shp = data.shape
-            other = [d for d in data_collection
-                     if d.shape == shp and d is not data]
-
-            # If no other datasets have the same shape, we go to the next one
-            if not other or skip_merge:
-                continue
-
-            if auto_merge:
-                merges, label = [data] + other, data.label
+            if all([data.shape == reference_shape for data in datasets[1:]]):
+                suggested_label = data_collection.suggest_merge_label(*datasets)
+                return data_collection.merge(*datasets, label=suggested_label)
             else:
-                merges, label = cls._choose_merge(data, other)
+                raise ValueError("Can only auto-merge datasets if they all have "
+                                 " the same shape")
 
-            if merges:
-                data_collection.merge(*merges, label=label)
+        else:
 
-            suggested.append(data)
-            suggested.extend(other)
+            return datasets
 
     @staticmethod
-    def _choose_merge(data, other):
+    def _choose_merge(data, other, suggested_label):
         """
         Present an interface to the user for approving or rejecting
         a proposed data merger. Returns a list of datasets from other
@@ -308,8 +334,8 @@ class Application(HubListener):
 
     @property
     def viewers(self):
-        """Return a tuple of tuples of viewers currently open
-        The i'th tuple stores the viewers in the i'th close_tab
+        """
+        Return a tuple of tuples of viewers currently open.
         """
         return []
 
@@ -341,175 +367,3 @@ class Application(HubListener):
                 viewer = context.object(v)
                 self.add_widget(viewer, tab=i, hold_position=True)
         return self
-
-
-class ViewerBase(HubListener, PropertySetMixin):
-
-    """ Base class for data viewers in an application """
-
-    # the glue.core.layer_artist.LayerArtistContainer
-    # class/subclass to use
-    _layer_artist_container_cls = None
-
-    def __init__(self, session):
-
-        HubListener.__init__(self)
-        PropertySetMixin.__init__(self)
-
-        self._session = session
-        self._data = session.data_collection
-        self._hub = None
-        self._layer_artist_container = self._layer_artist_container_cls()
-
-    def close(self):
-        self._layer_artist_container.clear()
-
-    def register_to_hub(self, hub):
-        self._hub = hub
-
-    def unregister(self, hub):
-        """ Abstract method to unsubscribe from messages """
-        raise NotImplementedError
-
-    def request_add_layer(self, layer):
-        """ Issue a command to add a layer """
-        cmd = command.AddLayer(layer=layer, viewer=self)
-        self._session.command_stack.do(cmd)
-
-    def add_layer(self, layer):
-        if isinstance(layer, Data):
-            self.add_data(layer)
-        elif isinstance(layer, Subset):
-            self.add_subset(layer)
-        # else: SubsetGroup
-
-    def add_data(self, data):
-        """ Add a data instance to the viewer
-
-        This must be overridden by a subclass
-
-        Parameters
-        ----------
-        data : :class:`~glue.core.data.Data`
-            Data object to add.
-        """
-        raise NotImplementedError
-
-    def add_subset(self, subset):
-        """ Add a subset to the viewer
-
-        This must be overridden by a subclass
-
-        Parameters
-        ----------
-        subset : :class:`~glue.core.subset.Subset`
-            Subset instance to add.
-        """
-        raise NotImplementedError
-
-    def apply_roi(self, roi):
-        """ Apply an ROI to the client
-
-        Parameters
-        ----------
-        roi : :class:`~glue.core.roi.Roi`
-            The ROI to apply.
-        """
-        raise NotImplementedError
-
-    @property
-    def session(self):
-        return self._session
-
-    @property
-    def axes(self):
-        return self.client.axes
-
-    def layer_view(self):
-        raise NotImplementedError()
-
-    def options_widget(self):
-        raise NotImplementedError()
-
-    def move(self, x=None, y=None):
-        """ Reposition a viewer within the application.
-
-        x : int, optional
-            Offset of viewer's left edge from the left edge of the parent
-            window.
-        y : int, optional
-            Offset of the viewer's top edge from the top edge of the parent
-            window.
-        """
-        raise NotImplementedError()
-
-    @property
-    def position(self):
-        """
-        Return the location of the viewer as a tuple of ``(x, y)``
-        """
-        raise NotImplementedError()
-
-    @property
-    def viewer_size(self):
-        """
-        Return the size of the viewer as a tuple of ``(width, height)``
-        """
-        raise NotImplementedError()
-
-    @viewer_size.setter
-    def viewer_size(self, value):
-        """ Resize the width and/or height of the viewer
-
-        Parameters
-        ----------
-        value : tuple of int
-            The width and height of the viewer.
-        width : int, optional
-            New width.
-        height : int, optional
-            New height.
-        """
-        raise NotImplementedError()
-
-    def restore_layers(self, rec, context):
-        """
-        Given a list of glue-serialized layers, restore them
-        to the viewer
-        """
-        # if this viewer manages a client, rely on it to restore layers
-        if hasattr(self, 'client'):
-            return self.client.restore_layers(rec, context)
-        raise NotImplementedError()
-
-    @property
-    def layers(self):
-        """Return a tuple of layers in this viewer.
-
-        A layer is a visual representation of a dataset or subset within
-        the viewer"""
-        return tuple(self._layer_artist_container)
-
-    def __gluestate__(self, context):
-        return dict(session=context.id(self._session),
-                    size=self.viewer_size,
-                    pos=self.position,
-                    properties=dict((k, context.id(v))
-                                    for k, v in self.properties.items()),
-                    layers=list(map(context.do, self.layers))
-                    )
-
-    @classmethod
-    def __setgluestate__(cls, rec, context):
-        session = context.object(rec['session'])
-        result = cls(session)
-        result.register_to_hub(session.hub)
-        result.viewer_size = rec['size']
-        x, y = rec['pos']
-        result.move(x=x, y=y)
-
-        prop = dict((k, context.object(v)) for
-                    k, v in rec['properties'].items())
-        result.restore_layers(rec['layers'], context)
-        result.properties = prop
-        return result
