@@ -19,6 +19,8 @@ from __future__ import absolute_import, division, print_function
 
 import logging
 
+import numpy as np
+
 from glue.external import six
 from glue.core.hub import HubListener
 from glue.core.message import DataCollectionDeleteMessage, DataRemoveComponentMessage
@@ -29,9 +31,10 @@ from glue.core.data import Data
 from glue.core.component import DerivedComponent
 from glue.core.exceptions import IncompatibleAttribute
 from glue.core.subset import Subset
+from glue.utils import unbroadcast
 
 __all__ = ['accessible_links', 'discover_links', 'find_dependents',
-           'LinkManager', 'is_equivalent_cid']
+           'LinkManager', 'is_equivalent_cid', 'pixel_cid_to_pixel_cid_matrix']
 
 
 def accessible_links(cids, links):
@@ -67,7 +70,7 @@ def discover_links(data, links):
     # TODO: try to add shortest paths first -- should
     # prevent lots of repeated checking
 
-    cids = set(data.primary_components)
+    cids = set(data.main_components + data.coordinate_components)
     cid_links = {}
     depth = {}
     for cid in cids:
@@ -133,7 +136,7 @@ class LinkManager(HubListener):
     """
 
     def __init__(self, data_collection=None):
-        self._external_links = set()
+        self._external_links = []
         self.hub = None
         self.trigger = False
         self.data_collection = data_collection
@@ -149,27 +152,23 @@ class LinkManager(HubListener):
         remove = []
         remove_cid = msg.component_id
         for link in self._external_links:
-            ids = set(link.get_from_ids()) | set([link.get_to_id()])
-            for cid in ids:
-                if cid is remove_cid:
-                    remove.append(link)
-                    break
+            if remove_cid in link:
+                remove.append(link)
         for link in remove:
             self.remove_link(link)
 
     def _data_removed(self, msg):
         remove = []
         for link in self._external_links:
-            ids = set(link.get_from_ids()) | set([link.get_to_id()])
-            for cid in ids:
-                if cid.parent is msg.data:
+            for cid in msg.data.components:
+                if cid in link and cid.parent is msg.data:
                     remove.append(link)
                     break
         for link in remove:
             self.remove_link(link)
 
     def clear_links(self):
-        self._external_links.clear()
+        self._external_links[:] = []
 
     def add_link(self, link, update_external=True):
         """
@@ -180,20 +179,20 @@ class LinkManager(HubListener):
         link : ComponentLink, LinkCollection, or list thereof
            The link(s) to ingest
         """
-        if isinstance(link, (LinkCollection, list)):
+        if isinstance(link, list):
             for l in link:
                 self.add_link(l, update_external=False)
             if update_external:
                 self.update_externally_derivable_components()
         else:
-            if link.inverse not in self._external_links:
-                self._external_links.add(link)
+            if link not in self._external_links and isinstance(link, LinkCollection) or link.inverse not in self._external_links:
+                self._external_links.append(link)
                 if update_external:
                     self.update_externally_derivable_components()
 
     @contract(link=ComponentLink)
     def remove_link(self, link, update_external=True):
-        if isinstance(link, (LinkCollection, list)):
+        if isinstance(link, list):
             for l in link:
                 self.remove_link(l, update_external=False)
             if update_external:
@@ -233,6 +232,9 @@ class LinkManager(HubListener):
         else:
             data_collection = self.data_collection
 
+        # Only keep actual Data instances since only they support links for now
+        data_collection = [d for d in data_collection if isinstance(d, Data)]
+
         for data in data_collection:
             links = discover_links(data, self._links | self._inverse_links)
             comps = {}
@@ -253,11 +255,22 @@ class LinkManager(HubListener):
 
     @property
     def _links(self):
+
         if self.data_collection is None:
             data_links = set()
         else:
-            data_links = set(link for data in self.data_collection for link in data.links)
-        return data_links | self._external_links
+            data_links = set(link for data in self.data_collection for link in getattr(data, 'links', []))
+
+        external_links = set()
+
+        for link in self._external_links:
+            if isinstance(link, LinkCollection):
+                for sublink in link:
+                    external_links.add(sublink)
+            else:
+                external_links.add(link)
+
+        return data_links | external_links
 
     @property
     def _inverse_links(self):
@@ -269,10 +282,10 @@ class LinkManager(HubListener):
 
     @property
     def external_links(self):
-        return list(self._external_links)
+        return self._external_links
 
     def clear(self):
-        self._external_links.clear()
+        self._external_links[:] = []
 
     def __contains__(self, item):
         return item in self._links
@@ -305,7 +318,7 @@ def is_equivalent_cid(data, cid1, cid2):
 
     Parameters
     ----------
-    data : `~glue.core.Data`
+    data : `~glue.core.data.Data`
         The data object in which to check for the component IDs
     cid1, cid2 : `~glue.core.ComponentID`
         The two component IDs to compare
@@ -327,7 +340,7 @@ def is_convertible_to_single_pixel_cid(data, cid):
 
     Parameters
     ----------
-    data : `~glue.core.Data`
+    data : `~glue.core.data.Data`
         The data in which to check for pixel components IDs
     cid : `~glue.core.ComponentID`
         The component ID to search for
@@ -389,3 +402,26 @@ def equivalent_pixel_cids(reference, target):
         else:
             return None
     return order
+
+
+def pixel_cid_to_pixel_cid_matrix(data1, data2):
+    """
+    Given two datasets, return a boolean matrix indicating which of the pixel
+    components are linked.
+
+    The returned matrix has the shape (data1.ndim, data2.ndim).
+    """
+
+    # For simplicity, and to avoid code used elsewhere, we rely on the fact
+    # that independent coordinates are broadcasted during conversions.
+
+    matrix = np.zeros((data1.ndim, data2.ndim), dtype=bool)
+
+    for idim, pix_cid in enumerate(data1.pixel_component_ids):
+        try:
+            pix_coords = unbroadcast(data2[pix_cid, [slice(2)] * data2.ndim])
+            matrix[idim] = np.array(pix_coords.shape) == 2
+        except IncompatibleAttribute:
+            pass
+
+    return matrix
