@@ -9,7 +9,7 @@ class GlueItemWidget(object):
 
     """
     A mixin for QtWidgets.QListWidget/GlueTreeWidget subclasses, that provides
-    drag+drop funtionality.
+    drag+drop functionality.
     """
     # Implementation detail: QXXWidgetItems are unhashable in PySide,
     # and cannot be used as dictionary keys. we hash on IDs instead
@@ -65,7 +65,7 @@ class GlueItemWidget(object):
         """
         Convenience method to set data associated with a ``QxxWidgetItem``.
         """
-        #item.setData(Qt.UserRole, data)
+        # item.setData(Qt.UserRole, data)
         self._mime_data[id(item)] = data
 
     def drop_data(self, item):
